/*
 * Copyright (c) 2013, Victor Nazarov <asviraspossible@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Victor Nazarov nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.github.sviperll;

public class Functions {
    @SuppressWarnings("rawtypes")
    private static Function ID = new IdFunction();

    @SuppressWarnings("unchecked")
    public static <T> ConfigurableFunction<T, T> id() {
        return new ConfigurableFunction<T, T>((Function<T, T>)ID);
    }

    public static <T, Q, R> ConfigurableFunction<T, R> compose(final Function<T, Q> f, final Function<? super Q, R> g) {
        return new ConfigurableFunction<T, R>(new ComposeFunction<T, Q, R>(f, g));
    }

    public static <T, R> ConfigurableFunction<T, R> passNullThrough(final Function<T, R> function) {
        return new ConfigurableFunction<T, R>(new PassNullThroughFunction<T, R>(function));
    }

    public static <T, R> ConfigurableFunction<T, R> throwOnNull(final Function<T, R> function) {
        return new ConfigurableFunction<T, R>(new ThrowOnNullFunction<T, R>(function));
    }

    public static <T, R> ConfigurableFunction<T, R> configure(Function<T, R> function) {
        if (function instanceof ConfigurableFunction)
            return (ConfigurableFunction<T, R>)function;
        else
            return new ConfigurableFunction<T, R>(function);
    }

    private Functions() {}

    private static class IdFunction<T> implements Function<T, T> {
        @Override
        public T apply(T t) {
            return t;
        }
    }

    private static class ComposeFunction<T, Q, R> implements Function<T, R> {
        private final Function<? super T, Q> f;
        private final Function<? super Q, R> g;
        public ComposeFunction(Function<? super T, Q> f, Function<? super Q, R> g) {
            this.f = f;
            this.g = g;
        }

        @Override
        public R apply(T t) {
            return g.apply(f.apply(t));
        }
    }

    private static class PassNullThroughFunction<T, R> implements Function<T, R> {
        private final Function<T, R> function;
        public PassNullThroughFunction(Function<T, R> function) {
            this.function = function;
        }

        @Override
        public R apply(T t) {
            if (t == null)
                return null;
            else
                return function.apply(t);
        }
    }

    private static class ThrowOnNullFunction<T, R> implements Function<T, R> {
        private final Function<T, R> function;
        public ThrowOnNullFunction(Function<T, R> function) {
            this.function = function;
        }

        @Override
        public R apply(T t) {
            if (t == null)
                throw new NullPointerException();
            else
                return function.apply(t);
        }
    }
}
