/*
 * Copyright (C) 2013 Victor Nazarov <asviraspossible@gmail.com>
 */

package com.github.sviperll;

public class Isomorphisms {
    @SuppressWarnings("rawtypes")
    private static final IdentityIsomorphism IDENTITY = new IdentityIsomorphism();

    public static <T, U> ConfigurableIsomorphism<T, U> configure(Isomorphism<T, U> isomorphism) {
        if (isomorphism instanceof ConfigurableIsomorphism)
            return (ConfigurableIsomorphism<T, U>)isomorphism;
        else
            return new ConfigurableIsomorphism<T, U>(isomorphism);
    }

    @SuppressWarnings("unchecked")
    public static <T> ConfigurableIsomorphism<T, T> identity() {
        return new ConfigurableIsomorphism<T, T>((IdentityIsomorphism<T>)IDENTITY);
    }

    public static <T, U, V> ConfigurableIsomorphism<T, V> compose(final Isomorphism<T, U> isomorphism1, final Isomorphism<U, V> isomorphism2) {
        return new ConfigurableIsomorphism<T, V>(new Isomorphism<T, V>() {
            @Override
            public V forward(T object) {
                return isomorphism2.forward(isomorphism1.forward(object));
            }

            @Override
            public T backward(V object) {
                return isomorphism1.backward(isomorphism2.backward(object));
            }
        });
    }

    public static <T, U> ConfigurableIsomorphism<U, T> reversed(final Isomorphism<T, U> isomorphism) {
        return new ConfigurableIsomorphism<U, T>(new Isomorphism<U, T>() {

            @Override
            public T forward(U object) {
                return isomorphism.backward(object);
            }

            @Override
            public U backward(T object) {
                return isomorphism.forward(object);
            }
        });
    }

    public static <T, U> ConfigurableIsomorphism<T, U> passNullThrough(final Isomorphism<T, U> isomorphism) {
        return new ConfigurableIsomorphism<T, U>(new Isomorphism<T, U>() {

            @Override
            public U forward(T object) {
                if (object == null)
                    return null;
                else
                    return isomorphism.forward(object);
            }

            @Override
            public T backward(U object) {
                if (object == null)
                    return null;
                else
                    return isomorphism.backward(object);
            }
        });
    }

    public static <T, U> ConfigurableIsomorphism<T, U> throwOnNull(final Isomorphism<T, U> isomorphism) {
        return new ConfigurableIsomorphism<T, U>(new Isomorphism<T, U>() {

            @Override
            public U forward(T object) {
                if (object == null)
                    throw new NullPointerException();
                else
                    return isomorphism.forward(object);
            }

            @Override
            public T backward(U object) {
                if (object == null)
                    throw new NullPointerException();
                else
                    return isomorphism.backward(object);
            }
        });
    }

    public static <T, U> ConfigurableFunction<T, U> asForwardFunction(final Isomorphism<T, U> isomorphism) {
        return new ConfigurableFunction<T, U>(new Function<T, U>() {
            @Override
            public U apply(T t) {
                return isomorphism.forward(t);
            }
        });
    }

    public static <T, U> ConfigurableFunction<U, T> asBackwardFunction(final Isomorphism<T, U> isomorphism) {
        return new ConfigurableFunction<U, T>(new Function<U, T>() {

            @Override
            public T apply(U t) {
                return isomorphism.backward(t);
            }
        });
    }

    private Isomorphisms() {
    }

    private static class IdentityIsomorphism<T> implements Isomorphism<T, T> {
        @Override
        public T forward(T object) {
            return object;
        }

        @Override
        public T backward(T object) {
            return object;
        }
    }
}
